from typing import Dict, List, Tuple


class Function(object):
    """
    Function を表すクラス。
    必要な情報を詰め込み、 to_pybind_string で生成する。
    """

    def __init__(self):
        self._return_type: str = ""
        self._arguments: List[Tuple[str, str]] = []
        self._name: str | None = None
        self._full_name: str | None = None
        self._namespace: List[str] = []
        self._description = ""
        self._module: str | None = None
        self._pyname: str | None = None

    def set_name(self, name: str, namespace: List[str]):
        self._name = name
        self._namespace = namespace
        self._full_name = f"{'::'.join(namespace)}::{name}"

    def set_return_type(self, type: str):
        self._return_type = type

    def set_argument_types(self, types: List[Tuple[str, str]]):
        """
        parameter: [(name, type),]
        """
        self._arguments = types

    @property
    def pyname(self):
        self._pyname

    @pyname.setter
    def pyname(self, python_name: str):
        self._pyname = python_name

    def add_argument_type(self, type: Tuple[str, str]):
        """
        parameter: (name, type)
        """
        self._arguments.append(type)

    def set_description(self, description: str):
        self._description = description

    def set_module(self, module: str):
        self._module = module

    def to_pybind_string(self):
        if self._name == None or self._full_name == None or self._module == None:
            print("Parse Error Skipping ...")
            return ""
        args = [f', pybind11::arg("{i[0]}")' for i in self._arguments]
        if self._pyname is not None:
            return (
                f'{self._module}.def("{self._pyname}", &{self._full_name}, "{self._description}"'
                f'{"".join(args)});'
            )
        else:
            return (
                f'{self._module}.def("{self._name}", &{self._full_name}, "{self._description}"'
                f'{"".join(args)});'
            )

    def to_decl_string(self):
        if self._name == None or self._full_name == None or self._module == None:
            print("Parse Error Skipping ...")
            return ""
        args = [f"{i[1]} {i[0]}" for i in self._arguments]
        return (
            f'namespace {"::".join(self._namespace)} '
            f'{{ {self._return_type} {self._name}({", ".join(args)}); }}'
        )

    def __eq__(self, obj):
        if isinstance(obj, Function):
            return self._full_name == obj._full_name
        else:
            return False


class StructOrClass:
    """
    Struct や Class を表すクラス
    必要な情報を詰め込み、 to_pybind_string で生成する。
    """

    def __init__(self):
        self._name: str | None = None
        self._namespace: List[str] = []
        self._members: list[Dict[str, bool | str]] = []
        self._member_funcs: list[Dict[str, bool | str | List[Tuple[str, str]]]] = []
        self._module: str | None = None
        self._description = ""

    def set_name(self, name: str, namespace: List[str]):
        self._name = name
        self._namespace = namespace
        self._full_name = f"{'::'.join(namespace)}::{name}"

    def add_member(
        self, name: str, type: str, description: str = "", private: bool = False
    ):
        self._members.append(
            {
                "name": name,
                "type": type,
                "description": description,
                "private": private,
            }
        )

    def add_member_func(
        self,
        name: str,
        type: str,
        args: List[Tuple[str, str]],  # list[(name, type)]
        description: str = "",
        private: bool = False,
    ):
        self._member_funcs.append(
            {
                "name": name,
                "return_type": type,
                "description": description,
                "private": private,
                "args": args,
            }
        )

    def set_module(self, module: str):
        self._module = module

    def set_description(self, description: str):
        self._description = description

    def get_members(self):
        return self._members

    def get_member_funcs(self):
        return self._member_funcs

    def to_pybind_string(self):
        if self._name == None or self._module == None:
            print("Parse Error Skipping ...")
            return ""
        return (
            f'pybind11::class_<::{self._full_name}>({self._module}, "{self._name}")\n'
            "\t\t.def(pybind11::init())"
            ## Member 変数の宣言
            + "\n".join(
                [""]
                + [
                    f'\t\t.def_readwrite("{i["name"]}",'
                    f' &{self._full_name}::{i["name"]}, "{i["description"]}")'
                    for i in self._members
                    if not i["private"]
                ]
            )
            ## Member 関数の宣言
            + "\n".join(
                [""]
                + [
                    f'\t\t.def("{i["name"]}",'
                    f' &{self._full_name}::{i["name"]}, "{i["description"]}")'
                    for i in self._member_funcs
                    if not i["private"]
                ]
            )
            + ";"
        )


class Submodule:
    """
    Submodule を表すクラス
    必要な情報を詰め込み、 to_pybind_string で生成する。
    """

    def __init__(self):
        self._name: str | None = None
        self._description = ""
        self._parents: List[str] = []

    @property
    def cpp_name(self) -> str:
        if self._name == None:
            print("Parse Error Skipping ...")
            return ""
        return "_".join(self._parents) + "_" + self._name

    @property
    def cpp_parent_name(self) -> str:
        if self._name == None:
            print("Parse Error Skipping ...")
            return ""
        return "_".join(self._parents)

    def set_name(self, name: str):
        self._name = name

    def set_description(self, description: str):
        self._description = description

    def set_parent(self, parents: List[str]):
        self._parents = parents

    def to_pybind_string(self):
        if self._name == None:
            print("Parse Error Skipping ...")
            return ""
        return f'auto {self.cpp_name} = {self.cpp_parent_name}.def_submodule("{self._name}", "{self._description}");'

    def __eq__(self, obj):
        if isinstance(obj, Submodule):
            return self._name == obj._name
        else:
            return False



import copy
import os
import re
from typing import List

from clang.cindex import AccessSpecifier, Config, Cursor, CursorKind, TranslationUnit

from .component import Function, StructOrClass, Submodule
from .logging import get_logger

logger = get_logger("parser")


class Parser:
    """
    CPPyGen はヘッダーを解析し、pybind11 用の関数を自動で作り出すコードジェネレー
    タです。
    """

    def __init__(
        self,
        namespace: str | None = None,
        *,
        library_path: str | None = None,
        library_file: str | None = None,
    ):
        self._funcitons: List[Function] = []
        self._submodules: List[Submodule] = []
        self._structs_and_classes: List[StructOrClass] = []
        self._hpp_includes: List[str] = []
        self._namespace = namespace or "cppygen"

        if library_file != None and library_path != None:
            raise ValueError(f"Both library_path and library_file cannot be set.")
        if not library_path is None:
            Config.set_library_path(library_path)
            return
        if not library_file is None:
            Config.set_library_file(library_file)
            return
        if (library_file is None and library_path is None) and (
            cppygen_libclang_path := os.environ.get("CPPYGEN_LIBCLANG_PATH", None)
        ) is not None:
            Config.set_library_file(cppygen_libclang_path)
            return

    def _get_tu(self, source: str, filename: str, flags=[]) -> TranslationUnit:
        if flags == None:
            flags = []
        if (cppygen_flags := os.environ.get("CPPYGEN_COMPILE_FLAGS", None)) is not None:
            flags.extend(cppygen_flags.split(" "))
            print(cppygen_flags.split(" "))
        args = list(flags)
        name = filename
        return TranslationUnit.from_source(
            name,
            args,
            unsaved_files=[(name, source)],
            options=TranslationUnit.PARSE_DETAILED_PROCESSING_RECORD,
        )

    def _extract_functions(self, cu: Cursor, namespace: List[str], module_name: str):
        """
        cu 以下にある関数を抽出する
        """
        for i in list(cu.get_children()):
            i: Cursor
            if i.kind == CursorKind.FUNCTION_DECL and i.is_definition():  # type: ignore
                func = Function()
                func.set_return_type(i.result_type.spelling)
                func.set_name(i.spelling, namespace)
                func.set_module(module_name)

                # extract comment string
                raw_comment = i.raw_comment or ""
                pyname = re.search(r"pyname: *(.*)", raw_comment)
                description = re.search(r"description: *(.*)", raw_comment)
                if pyname is not None:
                    func.pyname = pyname[1]
                func.set_description(description and description[1] or "")

                for j in list(i.get_children()):
                    j: Cursor
                    if j.kind == CursorKind.PARM_DECL:  # type: ignore
                        func.add_argument_type((j.spelling, j.type.spelling))
                self._funcitons.append(func)

    def _extract_struct_and_class(
        self, cu: Cursor, namespace: List[str], module_name: str
    ):
        """
        cu 以下にある構造体を抽出する
        """
        for i in list(cu.get_children()):
            i: Cursor
            if i.kind == CursorKind.STRUCT_DECL or i.kind == CursorKind.CLASS_DECL:  # type: ignore
                struct_or_class = StructOrClass()
                struct_or_class.set_name(i.spelling, namespace)
                struct_or_class.set_module(module_name)
                struct_or_class.set_description(i.brief_comment or "")
                for j in list(i.get_children()):
                    j: Cursor
                    if j.kind == CursorKind.FIELD_DECL:  # type: ignore
                        # メンバー変数の抽出
                        struct_or_class.add_member(
                            j.spelling,
                            j.type.spelling,
                            j.brief_comment or "",
                            j.access_specifier == AccessSpecifier.PRIVATE,  # type: ignore
                        )
                    elif j.kind == CursorKind.CXX_METHOD:  # type: ignore
                        # メンバー関数の抽出
                        args = []
                        for k in list(j.get_children()):
                            if k.kind == CursorKind.PARM_DECL:  # type: ignore
                                args.append((k.spelling, k.type.spelling))
                        struct_or_class.add_member_func(
                            j.spelling,
                            j.result_type.spelling,
                            args,
                            j.brief_comment or "",
                            j.access_specifier == AccessSpecifier.PRIVATE,  # type: ignore
                        )
                self._structs_and_classes.append(struct_or_class)

    def add_hpp_includes(self, hpp: str):
        self._hpp_includes.append(hpp)

    def parse(
        self,
        source: str,
        filename: str,
        lang: str = "cpp",
        flags=[],
        with_diagnostic=False,
    ):
        tu: TranslationUnit = self._get_tu(source, filename, flags)
        if with_diagnostic:
            has_error = False
            for diag in tu.diagnostics:
                if diag.severity in [diag.Fatal, diag.Error]:
                    has_error = True
                    logger.error(
                        f"{diag.location.file}:{diag.location.line}:{diag.location.column}: warining: {diag.spelling} [{diag.option}]"
                    )
                if diag.severity == diag.Warning:
                    logger.warn(
                        f"{diag.location.file}:{diag.location.line}:{diag.location.column}: warining: {diag.spelling} [{diag.option}]"
                    )
            if has_error:
                exit(1)
        root: Cursor = tu.cursor
        for i in list(root.get_children()):
            i: Cursor

            # 再起的に関数を抽出する。
            def visit(x: Cursor, namespace: List[str], module_name: str):
                if lang == "cpp":
                    self._extract_functions(x, namespace, module_name)
                elif lang == "hpp":  # ヘッダーでのみクラスを抽出する。
                    self._extract_struct_and_class(x, namespace, module_name)
                for i in list(x.get_children()):
                    i: Cursor
                    namespace_in = copy.copy(namespace)
                    if i.kind == CursorKind.NAMESPACE:  # type: ignore
                        submod = Submodule()
                        submod.set_name(i.spelling)
                        submod.set_description(i.brief_comment or "")
                        submod.set_parent(copy.copy(namespace_in))
                        if not submod in self._submodules:
                            self._submodules.append(submod)
                        namespace_in.append(i.spelling)
                        visit(i, namespace_in, submod.cpp_name)

            # トップレベルの namespace を探す
            if i.kind == CursorKind.NAMESPACE and i.spelling == self._namespace:  # type: ignore
                visit(i, [self._namespace], self._namespace)

    def parse_from_file(self, filename: str, lang: str = "cpp", flags=[]):
        with open(filename, "r") as f:
            data = f.read()
        self.parse(data, filename, lang, flags, with_diagnostic=True)

    def to_decl_string(self):
        return (
            "/* Function Declarations Start */\n"
            + "\n".join([i.to_decl_string() for i in self._funcitons] + [""])
            + "/* Function Declarations End */\n\n"
        )

    def to_submod_string(self):
        return (
            "\t/* Submodule Declarations Start */\n"
            + "\n".join(["\t" + i.to_pybind_string() for i in self._submodules] + [""])
            + "\t/* Submodule Declarations End */\n\n"
        )

    def to_export_string(self):
        return (
            "\t/* Function Export Start */\n"
            + "\n".join(["\t" + i.to_pybind_string() for i in self._funcitons] + [""])
            + "\t/* Function Export End */\n\n"
            "\t/* Structs and Classes Export Start */\n"
            + "\n".join(
                ["\t" + i.to_pybind_string() for i in self._structs_and_classes] + [""]
            )
            + "\t/* Structs and Classes Export End */\n\n"
        )

    def generate(self) -> str:
        """
        inline 関数を実装した、 header only なコードを自動生成する関数
        """
        return (
            "#pragma once\n\n"
            "#include <pybind11/cast.h>\n"
            "#include <pybind11/pybind11.h>\n"
            "#include <pybind11/pytypes.h>\n"
            "#include <pybind11/stl.h>\n\n"
            "/* Custom Header Include Start */\n"
            + "\n".join([f'#include "{i}"' for i in self._hpp_includes])
            + "/* Custom Header Include End */\n\n"
            f"{self.to_decl_string()}\n"
            "\n"
            "namespace CPPyGen {\n\n"
            f"static inline void CPPyGenExport(pybind11::module_ {self._namespace})\n"
            "{\n\n"
            f"{self.to_submod_string()}\n\n"
            f"{self.to_export_string()}\n\n"
            "};\n"
            "}"
        )

    def cpp_generate(self) -> str:
        """
        hpp_generate と対になる。 Export の関数の実装部分を自動生成するコード
        """
        return (
            '#include "cppygen_generated.hpp"\n\n'
            "#include <pybind11/cast.h>\n"
            "#include <pybind11/pybind11.h>\n"
            "#include <pybind11/pytypes.h>\n"
            "#include <pybind11/stl.h>\n\n"
            "\n"
            f"{self.to_decl_string()}\n"
            "\n"
            "namespace CPPyGen {\n\n"
            f"void CPPyGenExport(pybind11::module_ {self._namespace})\n"
            "{\n\n"
            f"{self.to_submod_string()}\n\n"
            f"{self.to_export_string()}\n\n"
            "};\n"
            "}"
        )

    def hpp_generate(self) -> str:
        """
        Header で関数を宣言したりし実際に利用しやすいコードにする。
        """
        return (
            "#pragma once\n\n"
            "#include <pybind11/cast.h>\n"
            "#include <pybind11/pybind11.h>\n"
            "#include <pybind11/pytypes.h>\n"
            "#include <pybind11/stl.h>\n\n"
            "/* Custom Header Include Start */\n"
            + "\n".join([f'#include "{i}"' for i in self._hpp_includes])
            + "/* Custom Header Include End */\n\n"
            "namespace CPPyGen {\n\n"
            f"extern void CPPyGenExport(pybind11::module_ {self._namespace});\n\n"
            "}"
        )



import os
import sys
from logging import CRITICAL, ERROR, WARNING, Logger

import colorlog


def create_default_formatter() -> colorlog.ColoredFormatter:
    """Create a default formatter of log messages.

    This function is not supposed to be directly accessed by library users.
    """
    return colorlog.ColoredFormatter(
        "%(log_color)s%(message)s",
        no_color=False if _color_supported() else True,
    )


def _color_supported() -> bool:
    """Detection of color support."""
    # NO_COLOR environment variable:
    if os.environ.get("NO_COLOR", None):
        return False

    if not hasattr(sys.stderr, "isatty") or not sys.stderr.isatty():
        return False
    else:
        return True


def get_logger(name: str) -> Logger:
    handler = colorlog.StreamHandler()
    handler.setFormatter(create_default_formatter())
    logger = colorlog.getLogger(name)
    logger.addHandler(handler)
    return logger


__all__ = ["ERROR", "CRITICAL", "WARNING"]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cppygen']

package_data = \
{'': ['*']}

install_requires = \
['clang>=14.0,<15.0',
 'colorlog>=6.7.0,<7.0.0',
 'pytest>=7.1.3,<8.0.0',
 'toml>=0.10.2,<0.11.0']

entry_points = \
{'console_scripts': ['cppygen = cppygen.__main__:run']}

setup_kwargs = {
    'name': 'cppygen',
    'version': '0.1.2',
    'description': 'A simple c++ code generator for pybind11',
    'long_description': '# CPPYGEN\n\nAutomatic code generation for pybind11.\n\nPybind11 is a powerful library that exposes C++ types in Python and vice versa, \n\nThis generator will be generate c++ code for pybind11, and make it easy to\nwrite a python module using c++.\n\n## Installation\n```\npip install cppygen\n```\n\n## Usage Guide\n\nAfter installing cppygen, you can use `cppygen` command.\n\n```\ncppygen --config_file /path/to/cppygenconfig.toml --cwd /path/to/cwd\n```\n\nThis command will load config file, and parse C++ code and generate\nC++ pybind11 Code.\n\nAfter generating the code. Include the generated header to your program and\njust write in pybind11 manner. Be sure to link the generated cpp code.\n\n```cpp\nPYBIND11_MODULE(pyshell, m) {\n  CPPyGen::CPPyGenExport(m);\n}\n```\n\n### CMake\n\nUse `add_custom_command` to auto generate.\n\n```cmake\nset(cppygen_generated_hpp ${CMAKE_CURRENT_BINARY_DIR}/cppygen_generated.hpp)\nset(cppygen_generated_cpp ${CMAKE_CURRENT_BINARY_DIR}/cppygen_generated.cpp)\n\nfind_program(_CPPYGEN_GENERATOR cppygen)\n\nadd_custom_command(\n  OUTPUT ${cppygen_generated_hpp} ${cppygen_generated_cpp}\n  COMMAND\n    ${_CPPYGEN_GENERATOR} ARGS #\n    --config_file ${CMAKE_CURRENT_LIST_DIR}/cppygenconfig.toml #\n    --cwd ${CMAKE_CURRENT_LIST_DIR}\n  DEPENDS ${SHELL_SOURCES}\n  COMMENT\n    "Generating CPPyGen Code To ${cppygen_generated_hpp} and ${cppygen_generated_cpp}"\n  VERBATIM)\n```\n\n## Config\n`cppygen` command does not work without configuration.\nUse toml format configuration file.\n\n**sources** [array of path, **required**]\nPaths with `cppygen` will parse. `cppygen` can extract functions from\nsources.\n\n**headers** [array of path, **required**]\nPaths with `cppygen` will parse.`cppygen` can extract structs or classes from\nheaders.\n\n**output_dir** [path, **required**]\nOutput directory of generated code.\n\n**search_namespace** [string, optional]\nDefault is "cppygen", this option will define the namespace witch\nwill be parsed by `cppygen`. Outside of this namespace would be ignored.\n\n**include_headers** [array of filename, optional]\n`cppygen` does not resolve include paths, thus if you want to export C++\nclasses you should specify include filenames.\n\n**include_directories** [array of dir, optional]\nThese directories will be passed as absolute paths to parser include flags.\nSame as `flags =["-I/abs_path/to/dir"]`\n\n**flags** [array of string, optional]\nParser compile options.\n\n**libclang_path** [path, optional]\nPath to `libclang` shared library.\n\n## Examples\nSee the `examples` directry for sample projects.\n\n### Function\n```cpp\nnamespace cppygen {\n/**\n * pyname: py_foo\n * description: some description\n **/\nvoid foo() {\n\n}\n\n}\n```\nThis function will export to python as "py_foo".\n`description` would be python docstring.\n\n\n## Env\n\n```bash\nPYGEN_LIBCLANG_PATH # Path to clang shared library\nPYGEN_COMPILE_FLAGS # additional flags to parse file\n```\n',
    'author': 'Gen740',
    'author_email': 'keener_slimier_0m@icloud.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import argparse
import pathlib

from .logging import get_logger

logger = get_logger("cppygen command")

import toml

from cppygen.cppygen_parser import Parser


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--config_file", required=True, type=str, help="Path to config file"
    )

    parser.add_argument(
        "--cwd", required=True, type=str, help="Current Working Directory"
    )

    args = parser.parse_args()

    configs = toml.load(args.config_file)

    cwd = pathlib.Path(args.cwd)

    sources = []
    if configs["sources"] is None:
        logger.error("Please Specify the sources field in config file")
        exit(1)
    for i in configs["sources"]:
        sources.extend([j for j in cwd.glob(i)])

    headers = []
    if configs["headers"] is None:
        logger.error("Please Specify the headers field in config file")
        exit(1)
    for i in configs["headers"]:
        headers.extend([j for j in cwd.glob(i)])

    if configs["output_dir"] is None:
        logger.error("Please Specify the output_dir field in config file")
        exit(1)
    output_dir = cwd.joinpath(configs["output_dir"])

    cppygen = Parser(
        namespace=configs["search_namespace"], library_file=configs["libclang_path"]
    )

    flags = configs["flags"] or []

    for i in configs["include_directories"] or []:
        flags.append(f"-I{str(cwd.joinpath(i).absolute())}")
        print(f"-I{str(cwd.joinpath(i).absolute())}")

    for i in sources:
        cppygen.parse_from_file(i, lang="cpp", flags=configs["flags"])

    for i in headers:
        cppygen.parse_from_file(i, lang="hpp", flags=configs["flags"])

    for i in configs["include_headers"] or []:
        cppygen.add_hpp_includes(i)

    with open(str(output_dir) + "/cppygen_generated.hpp", "w") as f:
        f.write(cppygen.hpp_generate())

    with open(str(output_dir) + "/cppygen_generated.cpp", "w") as f:
        f.write(cppygen.cpp_generate())


if __name__ == "__main__":
    run()
